#ifndef _PDP_COMMON_H_INCLUDED_
#define _PDP_COMMON_H_INCLUDED_

/**
 * @file
 * общие функции для работы с метками безопасности
 */

#ifdef __cplusplus
extern "C" {
#endif


#ifdef __KERNEL__
	#include <linux/types.h>

	#define PDP_USRKRNL(u,k) k
	#define PDP_KRNL(x) x
	#define PDP_NKRNL(x)
	#define PDP_KRNL_ARG(x) , x
#else
	#include <inttypes.h> /* 4 uint*_t */
	#include <stdlib.h> /* 4 size_t */

	#define PDP_USRKRNL(u,k) u
	#define PDP_KRNL(x)
	#define PDP_NKRNL(x) x
	#define PDP_KRNL_ARG(x)
#endif

typedef uint8_t		 PDP_LEV_T;
#define PDP_PRNTF_LEV    PRIu8
#define PDP_SCNF_LEV     SCNu8

typedef uint32_t	PDP_ILEV_T;
#define PDP_PRNTF_ILEV    SCNu32
#define PDP_SCNF_ILEV     SCNu32
#define PDP_PRNTF_ILEV_HEX SCNx32

typedef uint8_t           PDP_ILINEAR_T;
#define PDP_PRNTF_ILINEAR SCNu8
#define PDP_SCNF_ILINEAR  SCNu8

typedef uint64_t	  PDP_CAT_T;
#define PDP_PRNTF_CAT     PRIx64
#define PDP_SCNF_CAT      SCNx64

typedef uint16_t	PDP_TYPE_T;
#define PDP_PRNTF_TYPE    PRIx16
#define PDP_SCNF_TYPE     SCNx16

typedef uint32_t	PDP_CNT_SIZE_T;
#define PDP_PRNTF_CNT_SIZE    PRIx32
#define PDP_SCNF_CNT_SIZE     SCNx32


#define PDPT_CCNR		0x01
#define PDPT_CCNRI		0x02
#define PDPT_EHOLE		0x04
#define PDPT_WHOLE		0x08

#ifdef __KERNEL__
	#define PDPL_PRINTFA(l) PDPML_PRINTFA(l->ml)

	#define PDPL_PRINTF  "%" "hhu"  ":%"  "hhu" ":%" "llx" ":%" "hx"
	#define PDPML_PRINTFA(ml) ml->lev,ml->ilev,ml->cat,ml->type
#else
	#define PDPL_PRINTFA(l) pdpl_lev(l),pdpl_ilev(l),pdpl_cat(l),pdpl_type(l)
	#define PDPL_PRINTF "%" PDP_PRNTF_LEV  ":%" PDP_PRNTF_ILEV ":%" PDP_PRNTF_CAT ":%" PDP_PRNTF_TYPE

	#define PDPML_PRINTFA(l) pdpml_lev(l),pdpml_ilev(l),pdpml_cat(l),pdpml_type(l)
#endif

	#define PDPML_PRINTF PDPL_PRINTF

/**
 * @brief Определение метки безопасности в соттветсвии с ДП моделью для объектов и субъектов доступа (PDP - Parsec Dostup Potoky)
 * Работа с экземплярами данного типа ведется только через указатель, внутреннее состояние доступно только через
 * вызов определенной функции (инкапсуляция)
 */
typedef struct PDP_LABEL_T			PDPL_T;

/**
 * @brief Определение мандатной метки в соттветсвии с ДП моделью для объектов и субъектов доступа (PDP - Parsec Dostup Potoky)
 * Работа с экземплярами данного типа ведется только через указатель, внутреннее состояние доступно только через
 * вызов определенной функции (инкапсуляция)
 */
typedef struct PDP_MAC_LABEL_T		PDPML_T;

#define PDPML_ZERO  (pdpl_ml(PDPL_ZERO))

extern const PDPL_T *PDPL_ZERO; /**< Нулевая метка безопасности: уровень конф - 0, целостности - низкий; категорий нет; флагов нет; ролей нет */
extern const PDPL_T *PDPL_EQU; /**< Метка безопасности объекта дырки: уровень конф - 0, целостности - низкий; категорий нет; флаг PDPT_EHOLE; ролей нет */
extern const PDPL_T *PDPL_MAXILEV; /**< Метка безопасности с высокой целосностью: уровень конф - 0, целостности - предварительно НЕОБХОДИМО УСТАНОВИТЬ; категорий нет; флагов нет; ролей нет */

/*@ requires \valid(l);
    requires \valid(&l->ucnt);
    assigns l->ucnt;
    ensures pdpl_incremented{Pre,Post}(l);
    ensures \result == l;
*/
const PDPL_T* pdpl_get(const PDPL_T *l);

/*@ requires \valid(l);
    assigns l->ucnt.counter;
    ensures l->ucnt.counter == \old(l->ucnt.counter) + 1;
    ensures \result == l;
 */
PDPL_T* pdplnc_get(PDPL_T *l);

/*@ requires \valid(ml);
    requires \valid(&ml->ucnt);
    assigns ml->ucnt;
    ensures pdpml_incremented{Pre,Post}(ml);
    ensures \result == ml;
*/
const PDPML_T* pdpml_get(const PDPML_T *ml);

/**
 * Выделение памяти и инициализация метки безопасности
 * @return указатель на метку, которая представляет собой копию нулевой метки (),
 * в случае ошибки выделения памяти NULL
 */
/*@ assigns \nothing;
    allocates \result;
    allocates \result->ml;
    allocates \result->rl;
    ensures \result == \null ||
       \valid(\result) && \offset_min(\result) == 0 &&
       (\result->ml == \null || pdpml_get_new_ensures(\result->ml)) &&
       (\result->rl == \null || pdprl_get_new_ensures(\result->rl)) &&
       \result->ucnt.counter == 1;
*/
PDPL_T* pdpl_get_new(PDP_USRKRNL(void,gfp_t f));

/*@ predicate pdpml_get_new_ensures(PDPML_T *pm) =
       \valid(pm) && \offset_min(pm) == 0 &&
       pm->lev  == 0 &&
       pm->ilev == 0 &&
       pm->cat  == 0 &&
       pm->type == 0 &&
       pm->catso64 == \null &&
       pm->catso64_size == 0 &&
       pm->ucnt.counter == 1;
*/
/*@ assigns \nothing;
    allocates \result;
    ensures \result == \null || pdpml_get_new_ensures(\result);
*/
PDPML_T* pdpml_get_new(PDP_USRKRNL(void,gfp_t f));

/* pdpXl_get_new_init */

/*@ assigns \nothing;
    allocates \result;
    allocates \result->ml;
    allocates \result->rl;
    ensures \result == \null || \valid(\result) && \offset_min(\result) == 0;
    ensures \valid(\result) ==>
            pdpml_get_new_init_ensures(\result->ml, lev, ilev, cat, type) &&
            (\result->rl == \null || pdprl_get_new_ensures(\result->rl)) &&
            \result->ucnt.counter == 1;
*/
PDPL_T* pdpl_get_new_init_mac(PDP_LEV_T lev, PDP_ILEV_T ilev, PDP_ILINEAR_T ilinear, PDP_CAT_T cat, PDP_TYPE_T type PDP_KRNL_ARG(gfp_t f));

/**
 * @brief Выделение памяти и инициализация мандатной метки безопасности
 *
 * @param lev
 * @param ilev
 * @param cat
 * @param type
 * @return указатель на метку, в случае ошибки выделения памяти NULL
 */
/*@ predicate pdpml_get_new_init_ensures(PDPML_T* pm, PDP_LEV_T lev, PDP_ILEV_T ilev, PDP_CAT_T cat, PDP_TYPE_T type) =
       \valid(pm) && \offset_min(pm) == 0 &&
       pm->lev          == lev   &&
       pm->ilev         == ilev  &&
       pm->cat          == cat   &&
       pm->type         == type  &&
       pm->catso64_size == 0     &&
       pm->catso64      == \null &&
       pm->ucnt.counter == 1;
*/
/*@ assigns \nothing;
    allocates \result;
    ensures \result == \null || pdpml_get_new_init_ensures(\result, lev, ilev, cat, type);
*/
PDPML_T* pdpml_get_new_init(PDP_LEV_T lev, PDP_ILEV_T ilev, PDP_ILINEAR_T ilinear, PDP_CAT_T cat, PDP_TYPE_T type PDP_KRNL_ARG(gfp_t f));

/*@ requires ml == \null || \valid(ml);
    requires rl == \null || \valid(rl);
    assigns \nothing;
    allocates \result;
    allocates \result->ml;
    allocates \result->rl;
    ensures \result == \null ||
       \valid(\result) && \offset_min(\result) == 0 &&
       \result->ucnt.counter == 1                   &&
       ml != \null ==> \result->ml == ml            &&
       rl != \null ==> \result->rl == rl            &&
       ml == \null ==> (\result->ml == \null || pdpml_get_new_ensures(\result->ml)) &&
       rl == \null ==> (\result->rl == \null || pdprl_get_new_ensures(\result->rl));
*/
PDPL_T* pdpl_get_new_init_mr(const PDPML_T *ml PDP_KRNL_ARG(gfp_t f));


/**
Освобождение ресурсов

@param l указатель на метку безопасности
*/

/*@ predicate pdpl_put_requires(PDPL_T* l) =
       pdpl_free_requires(l)                   &&
       \valid(&l->ucnt)                        &&
       l->ucnt.counter > 0;

    predicate pdpl_put_ensures{L1,L2}(PDPL_T* l) =
       ((\at(l->ucnt.counter,L1) == 1) ==> pdpl_free_ensures{L1,L2}(l)) &&
       ((\at(l->ucnt.counter,L1) > 1) ==> (\at(l->ucnt.counter,L2) == \at(l->ucnt.counter,L1) - 1));
*/
/*@ requires pdpl_put_requires(l);
    assigns l->ml->ucnt.counter;
    assigns l->rl->ucnt.counter;
    assigns l->ucnt.counter;
    ensures pdpl_put_ensures{Pre,Post}(l);
    behavior DEC:
       assumes l->ucnt.counter > 1;
       frees \nothing;
       ensures l->ucnt.counter == \old(l->ucnt.counter) - 1;
    behavior FREE:
       assumes l->ucnt.counter == 1;
       frees l->ml->catso64;
       frees l->ml;
       frees l->rl->aroles;
       frees l->rl->roles;
       frees l->rl;
       frees l;
       ensures pdpl_free_ensures{Pre,Post}(l);
    complete behaviors;
    disjoint behaviors;
*/
void pdpl_put(const PDPL_T* l);

/*@ predicate pdpml_put_requires(PDPML_T* ml) =
       \valid(&ml->ucnt)                        &&
       ml->ucnt.counter > 0;

    predicate pdpml_put_ensures{L1,L2}(PDPML_T* ml) =
       ((\at(ml->ucnt.counter,L1) == 1) ==> pdpml_free_ensures{L1,L2}(ml)) &&
       ((\at(ml->ucnt.counter,L1) > 1) ==> (\at(ml->ucnt.counter,L2) == \at(ml->ucnt.counter,L1) - 1));
*/
/*@ requires pdpml_put_requires(ml);
    assigns ml->ucnt.counter;
    ensures pdpml_put_ensures{Pre,Post}(ml);
    behavior DEC:
       assumes ml->ucnt.counter > 1;
       frees \nothing;
       ensures ml->ucnt.counter == \old(ml->ucnt.counter) - 1;
    behavior FREE:
       assumes ml->ucnt.counter == 1;
       frees ml->catso64;
       frees ml;
       ensures pdpml_free_ensures{Pre,Post}(ml);
    complete behaviors;
    disjoint behaviors;
*/
void pdpml_put(const PDPML_T* ml);

/**
 * @brief получение количества ссылок на метку безопасности, только для отладки
 * @param l указатель на исходную метку безопасности
 * @return количество ссылок на метку безопасности
*/
//@ logic integer pdpl_ucnt(PDPL_T *l) = l->ucnt.counter;
/*@ requires \valid(l);
    requires \valid(&l->ucnt);
    requires l->ucnt.counter >= 0;
    assigns \nothing;
    ensures \result == pdpl_ucnt(l);
 */
size_t pdpl_ucnt(const PDPL_T* l);

/*@ requires \valid(ml);
    requires ml->ucnt.counter >= 0;
    assigns \nothing;
    ensures \result == ml->ucnt.counter;
 */
size_t pdpml_ucnt(const PDPML_T* ml);


/**
 * @brief Дублирование метки безопасности
 * Создание ПОЛНОЙ копии исходной метки безопасности в динамически выделенной области памяти
 * @param l указатель на исходную метку безопасности
 * @return указатель на копию исходной метки безопасности, в случае ошибки выделения памяти NULL
*/

//@ ghost enum {SUCCESS, FAILURE} outcome_pdpl_dup;
/*@ requires \valid(l);
    requires
        (l->catso64 == \null && l->catso64_size == 0) ||
        \valid(((char *)l->catso64)+(0..l->catso64_size));
    requires l->aroles  == \null && l->aroles_cnt   == 0 || \valid(l->aroles);
    requires l->roles   == \null && l->roles_cnt    == 0 || \valid(l->roles);
    assigns outcome_pdpl_dup;
    behavior SUCCESS:
        ensures outcome_pdpl_dup == SUCCESS ==>
        \result != \null && \valid(\result) &&
        \result->lev  == l->lev  &&
        \result->ilev == l->ilev &&
        \result->cat  == l->cat  &&
        \result->type == l->type &&

        \result->catso64_size == l->catso64_size &&
        \result->aroles_cnt   == l->aroles_cnt   &&
        \result->roles_cnt    == l->roles_cnt    &&

        (l->catso64 == \null ==> \result->catso64 == \null) &&
        (l->aroles  == \null ==> \result->aroles  == \null) &&
        (l->roles   == \null ==> \result->roles   == \null) &&

        \result->ucnt.counter == 1;
    behavior FAILURE:
        ensures outcome_pdpl_dup == FAILURE ==> \result == \null;
    complete behaviors;
 */
PDPL_T* pdpl_dup(const PDPL_T* l PDP_KRNL_ARG(gfp_t f));

//@ ghost enum status pdpml_dup_alloc_status;
/*@ requires \valid(l);
    requires \typeof(l->catso64) <: \type(char *);
    requires l->catso64 == \null || \valid((char *)l->catso64+(0..l->catso64_size-1));
    assigns pdpml_dup_alloc_status;
    behavior ALLOC_OK:
       ensures pdpml_dup_alloc_status == SUCCESS ==>
               \valid(\result) && \offset_min(\result) == 0 &&
               pdpml_cmp(l, \result) &&
               \result->ucnt.counter == 1;
    behavior ALLOC_FAIL:
       ensures pdpml_dup_alloc_status == FAILURE ==>
               \result == \null;
    complete behaviors;
*/
PDPML_T* pdpml_dup(const PDPML_T* l PDP_KRNL_ARG(gfp_t f));


/**
 * @brief обновление метки безопасности
 * Создание ПОЛНОЙ копии исходной метки безопасности @src в @dst в соответсвии с концепцией RCU(D):
 *  - уменьшение счетчика ссылок метки *dsc (pdpl_put(*dsc);)
 *  - cоздание полной копии исходной метки *src (*dsc=pdpl_dup(src);)
 *
 * Результирующая метка гарантировано имеет ссчетчик ссылок равный 1;
 *
 * Данная операция потокоНЕбезопасная, требуется захват блокировки, сответвующей указателю @dst.
 *
 * @param src указатель на исходную метку безопасности
 * @param dst указатель на указатель на результирующую метку безопасности
 * @return в случае успеха *dsc, т.е. указатель на копию исходной метки безопасности;
 * в случае ошибки выделения памяти NULL: если src==dst errno выставляется в EINAVL, в случае ошибки выделения памяти - ENOMEM
*/
PDPL_T* pdpl_upd(PDPL_T **dst,const PDPL_T *src PDP_KRNL_ARG(gfp_t f) );


//@ logic PDPML_T* pdpl_ml(PDPL_T* l) = l->ml;
/*@ requires \valid(l);
    assigns \nothing;
    ensures \result == pdpl_ml(l);
 */
const PDPML_T* pdpl_ml(const PDPL_T* l);

//@ logic PDP_LEV_T pdpl_lev(PDPL_T* l) = pdpml_lev(pdpl_ml(l));
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns \nothing;
    ensures \result == pdpl_lev(l);
 */
PDP_LEV_T  pdpl_lev(const PDPL_T* l);
//@ logic PDP_ILEV_T pdpl_ilev(PDPL_T* l) = pdpml_ilev(pdpl_ml(l));
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns \nothing;
    ensures \result == pdpl_ilev(l);
 */
PDP_ILEV_T pdpl_ilev(const PDPL_T* l);
//@ logic PDP_CAT_T pdpl_cat{L}(PDPL_T* l) = pdpml_cat(pdpl_ml(l));
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns \nothing;
    ensures \result == pdpl_cat(l);
 */
PDP_CAT_T  pdpl_cat(const PDPL_T* l);
//@ logic PDP_TYPE_T pdpl_type(PDPL_T* l) = pdpml_type(pdpl_ml(l));
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns \nothing;
    ensures \result == pdpl_type(l);
 */
PDP_TYPE_T pdpl_type(const PDPL_T* l);

/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->lev;
    ensures pdpl_lev(l) == lev;
 */
void pdpl_set_lev(PDPL_T* l, PDP_LEV_T lev);
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->ilev;
    ensures pdpl_ilev(l) == ilev;
 */
void pdpl_set_ilev(PDPL_T* l, PDP_ILEV_T ilev);

/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->cat;
    ensures pdpl_cat(l) == cat;
 */
void pdpl_set_cat(PDPL_T* l, PDP_CAT_T cat);
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->cat;
    ensures pdpl_cat(l) == (pdpl_cat{Old}(l) & cat);
 */
void pdpl_and_cat(PDPL_T* l, PDP_CAT_T cat);
/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->cat;
    ensures pdpl_cat(l) == (pdpl_cat{Old}(l) | cat);
 */
void pdpl_or_cat(PDPL_T* l, PDP_CAT_T cat);

/*@ requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->type;
    ensures pdpl_type(l) == type;
 */
void pdpl_set_type(PDPL_T* l, PDP_TYPE_T type);
/*@ requires import_bitwise_integer_casts;
    requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->type;
    ensures pdpl_type(l) == (pdpl_type{Old}(l) & type);
 */
void pdpl_and_type(PDPL_T* l, PDP_TYPE_T type);
/*@ requires import_bitwise_integer_casts;
    requires \valid(l);
    requires \valid(l->ml);
    assigns l->ml->type;
    ensures pdpl_type(l) == (pdpl_type{Old}(l) | type);
 */
void pdpl_or_type(PDPL_T* l, PDP_TYPE_T type);

/*@ logic PDP_LEV_T pdpml_lev(PDPML_T* ml) = ml->lev;
 */
/*@ requires \valid(ml);
    assigns \nothing;
    ensures \result == pdpml_lev(ml);
 */
PDP_LEV_T  pdpml_lev(const PDPML_T* l);
/*@ logic PDP_ILEV_T pdpml_ilev(PDPML_T* ml) = ml->ilev;
 */
/*@ requires \valid(ml);
    assigns \nothing;
    ensures \result == pdpml_ilev(ml);
 */
PDP_ILEV_T pdpml_ilev(const PDPML_T* l);
/*@ logic PDP_CAT_T pdpml_cat{L}(PDPML_T* ml) = \at(ml->cat, L);
 */
/*@ requires \valid(ml);
    assigns \nothing;
    ensures \result == pdpml_cat(ml);
 */
PDP_CAT_T  pdpml_cat(const PDPML_T* l);
/*@ logic PDP_TYPE_T pdpml_type{L}(PDPML_T* ml) = \at(ml->type, L);
 */
/*@ requires \valid(ml);
    assigns \nothing;
    ensures \result == pdpml_type(ml);
 */
PDP_TYPE_T pdpml_type(const PDPML_T* l);

PDPL_T* pdpl_unite(PDPL_T* a,const PDPL_T* b, int flags); /* return a|b*/
PDPL_T* pdpl_subtract(PDPL_T* a, const PDPL_T* b, int flags); /* return a&b*/

/*@ requires \valid(ml);
    assigns ml->lev;
    ensures pdpml_lev(ml) == lev;
 */
void pdpml_set_lev(PDPML_T* l, PDP_LEV_T lev);
/*@ requires \valid(ml);
    assigns ml->ilev;
    ensures pdpml_ilev(ml) == ilev;
 */
void pdpml_set_ilev(PDPML_T* l, PDP_ILEV_T ilev);

/*@ requires \valid(ml);
    assigns ml->cat;
    ensures pdpml_cat(ml) == cat;
 */
void pdpml_set_cat(PDPML_T* l, PDP_CAT_T cat);
/*@ requires \valid(ml);
    assigns ml->cat;
    ensures pdpml_cat(ml) == (pdpml_cat{Old}(ml) & cat);
 */
void pdpml_and_cat(PDPML_T* l, PDP_CAT_T cat);
/*@ requires \valid(ml);
    assigns ml->cat;
    ensures pdpml_cat(ml) == (pdpml_cat{Old}(ml) | cat);
 */
void pdpml_or_cat(PDPML_T* l, PDP_CAT_T cat);

/*@ requires \valid(ml);
    assigns ml->type;
    ensures pdpml_type(ml) == type;
 */
void pdpml_set_type(PDPML_T* l, PDP_TYPE_T type);
/*@ requires import_bitwise_integer_casts;
    requires \valid(ml);
    assigns ml->type;
    ensures pdpml_type(ml) == (pdpml_type{Old}(ml) & type);
 */
void pdpml_and_type(PDPML_T* l, PDP_TYPE_T type);
/*@ requires \valid(ml);
    assigns ml->type;
    ensures pdpml_type(ml) == (pdpml_type{Old}(ml) | type);
 */
void pdpml_or_type(PDPML_T* l, PDP_TYPE_T type);

/**
Сравнение меток безопасности

@param a указатель на метку безопасности
@param b указатель на метку безопасности
@param flags указатель на метку безопасности

@return 1 - если равны, 0 - если не равны
*/
int pdpl_cmp(const PDPL_T* a, const PDPL_T* b, int flags);

/*@ predicate pdpml_cmp_requires(PDPML_T *a, PDPML_T *b) =
       (\valid(a) && \valid(b) || a == \null && b == \null) &&
       (\valid(a) ==> \typeof(a->catso64) <: \type(char *) && \valid(((char *)a->catso64)+(0..a->catso64_size - 1))) &&
       (\valid(b) ==> \typeof(b->catso64) <: \type(char *) && \valid(((char *)b->catso64)+(0..b->catso64_size - 1)));

    predicate pdpml_cmp(PDPML_T *a, PDPML_T *b) =
       a->lev          == b->lev          &&
       a->ilev         == b->ilev         &&
       a->cat          == b->cat          &&
       a->type         == b->type         &&
       a->catso64_size == b->catso64_size &&
       \forall integer i; 0 <= i < a->catso64_size ==> ((char*)a->catso64)[i] == ((char*)b->catso64)[i];
       //memcmp_char((char*)a->catso64, (char*)b->catso64, a->catso64_size);

    predicate pdpml_cmp_ensures(int res, PDPML_T *a, PDPML_T *b) =
       (res == 0 || res == 1) &&
       ((res == 1) ==> (a == b || pdpml_cmp(a, b))) &&
       ((res == 0) ==> (a != b && !pdpml_cmp(a, b)));
 */
/*@ requires pdpml_cmp_requires(a, b);
    assigns \nothing;
    ensures pdpml_cmp_ensures(\result, a, b);
    behavior PEQ:
       assumes a == b;
       ensures \result == 1;
    behavior EQ:
       assumes a != b;
       assumes pdpml_cmp(a, b);
       ensures \result == 1;
    behavior NE:
       assumes a != b;
       assumes !pdpml_cmp(a, b);
       ensures \result == 0;
    complete behaviors;
    disjoint behaviors;
*/
int pdpml_cmp(const PDPML_T* a, const PDPML_T* b, int flags);


/*@ requires intersectCategories;
    requires equalCategories;
    requires \valid(less);
    requires \valid(great);
    requires ValidIntLevel(less->ilev);
    requires ValidIntLevel(great->ilev);
    assigns \nothing;
    ensures \result == pdpml_notg(i2m_MacLabel(less), i2m_MacLabel(great));
*/
int pdpml_notg(const PDPML_T *less, const PDPML_T *great);
int pdpl_notg(const PDPL_T *less, const PDPL_T *great);

/**
Проверка, может ли объект-родитель с меткой @param parent иметь дочерний
объект с меткой @param child

@param child указатель на метку безопасности
@param parent указатель на метку безопасности

@return 0 если может, иначе не 0
*/
int pdp_child_parent(const PDPL_T *child, const PDPL_T *parent);

/*@ requires intersectCategories;
    requires equalCategories;
    requires \valid(child);
    requires \valid(parent);
    requires ValidIntLevel(child->ilev);
    requires ValidIntLevel(parent->ilev);
    assigns \nothing;
    ensures \result == pdpml_child_parent(i2m_MacLabel(child), i2m_MacLabel(parent));
*/
int pdpml_child_parent(const PDPML_T *child, const PDPML_T *parent);

/**
@breif Сравнение меток безопасности

Предосваляется для удобства, является надстройкой над функцией pdpl_cmp
static inline int pdpl_equ(const PDPL_T* a, const PDPL_T* b) { return 1 == pdpl_cmp(a,b,0); }

@param a указатель на метку безопасности
@param b указатель на метку безопасности
@return не 0 - если равны, 0 - если не равны
*/
static inline int pdpl_equ(const PDPL_T* a, const PDPL_T* b) { return 1 == pdpl_cmp(a, b, 0); }
static inline int pdpml_equ(const PDPML_T* a, const PDPML_T* b) { return 1==pdpml_cmp(a, b, 0); }

static inline int pdpl_is_zero(const PDPL_T* l) { return 1==pdpl_equ(l,PDPL_ZERO); }
static inline int pdpml_is_zero(const PDPML_T* ml) { return 1==pdpml_equ(ml,PDPML_ZERO); }

/**
Получение метки безопасности из бинарного представления, метка будет выделена посредсвом pdp_alloc

@param buf указатель на буфер, который содержти метку безопасности в бинарном виде
@param size размер буфера
@param __out size_without_role размер буфера, относящийся к манд. атрибутам
@return указатель на метку безопасности или NULL в случае ошибки (неверный формат метки)
*/
PDPL_T* pdpl_get_from_raw(const void *buf, size_t size PDP_KRNL_ARG(gfp_t f));

PDPML_T* pdpml_get_from_raw(const void *raw, size_t size, size_t* size_without_role PDP_KRNL_ARG(gfp_t f));


/**
Получение метки безопасности из бинарного представления

@param l указатель на метку безопасности
@param buf указатель на буфер, который содержти метку безопасности в бинарном виде
@param size размер буфера
@return 0 в случае успеха
*/
int pdpl_from_raw(PDPL_T* l, const void *buf, size_t size);

/**
Получение бинарного представления метки безопасности

@param l указатель на метку безопасности
@param size указатель на переменную, куда возвращается размер выделенного буфера (НЕ NULL)
@return указатель на буфер, выделенный функцией malloc(), который содержти метку безопасности в бинарном виде или NULL в случае ошибки
*/
void* pdpl_get_raw(const PDPL_T* l, size_t* size PDP_KRNL_ARG(gfp_t f)); /* буфер будет выделен функцией */
void* pdpml_get_raw(const PDPML_T* l, size_t* size PDP_KRNL_ARG(gfp_t f)); /* буфер будет выделен функцией */


extern const size_t PDPL_RAW_MIN_SIZE;

/**
Получение размера в байтах бинарного представления метки безопасности

@param l указатель на метку безопасности
@return размер в байтах бинарного представления метки безопасности
*/
/*@ logic integer pdpl_raw_size(PDPL_T* l) =
       PDPL_RAW_MIN_SIZE + l->ml->catso64_size +
       pdp_roles_raw_size(l->rl->roles_cnt) +
       pdp_roles_raw_size(l->rl->aroles_cnt);
*/
/*@ requires \valid(l);
    requires \valid(l->ml);
    requires \valid(l->rl);
    requires pdpl_raw_size(l) <= SIZE_MAX;
    assigns \nothing;
    ensures \result == pdpl_raw_size(l);
*/
size_t pdpl_raw_size(const PDPL_T* l);


/*@ logic integer pdpml_raw_size(PDPML_T *ml) =
       PDPLM_RAW_MIN_SIZE + ml->catso64_size;
 */
/*@ requires \valid(ml);
    requires pdpml_raw_size(ml) <= SIZE_MAX;
    assigns \nothing;
    ensures \result == pdpml_raw_size(ml);
*/
size_t pdpml_raw_size(const PDPML_T* ml);

/*@ logic size_t pdprawml_size(void* rawml) =
       (size_t)(sizeof(PDPML_RAW_H_T)+((PDPML_RAW_H_T*)rawml)->catso64_size);
 */
/*@ requires \typeof(rawml) <: \type(PDPML_RAW_H_T *);
    requires \valid((PDPML_RAW_H_T*)rawml);
    assigns \nothing;
    ensures \result == pdprawml_size(rawml);
 */
size_t pdprawml_size(const void* rawml);

typedef struct _parsec_mac_raw_t parsec_mac_raw_t;
void pdpl_2_legacy_raw(const PDPL_T* l, parsec_mac_raw_t* mr);
PDPL_T* pdpl_get_from_legacy_raw(parsec_mac_raw_t* mr PDP_KRNL_ARG(gfp_t f));

#ifdef __KERNEL__
	#define R_OK MAY_READ
	#define W_OK MAY_WRITE
	#define X_OK MAY_EXEC
	#define A_OK MAY_APPEND
	#define C_OK MAY_CREATE
#else
	#define A_OK (R_OK<<1) /* APPEND need description */
	#define C_OK (A_OK<<1)
#endif
	#define PDP_O_CONTAINER   0x08000
	#define LEGACY_IGNORE_LEV 0x10000
	#define LEGACY_IGNORE_CAT 0x20000
	#define PDP_IGN_INT       0x40000

/*
 * Проверка доступа объека с субъекту
 *
 * @param o указатель на метку безопасности объекта
 * @param s указатель на метку безопасности субъекта
 * @param mode режим доступа X_OK, R_OK, W_OK
 *
 * @return 0 - доступ разрешен, иначе - запрещен
*/

/* mode: X_OK, R_OK, W_OK or MAY_READ, MAY_WRITE, MAY_EXEC 4 kernel */
/*@ requires \valid(s);
    requires \valid(o);
    requires \valid(s->ml);
    requires \valid(o->ml);
    requires ValidIntLevel(s->ml->ilev);
    requires ValidIntLevel(o->ml->ilev);
    assigns \nothing;
    ensures \result == (int) pdpml_permission(i2m_MacLabel(s->ml), i2m_MacLabel(o->ml), mode);
*/
int pdpl_permission(const PDPL_T* s, const PDPL_T* o, int mode);

/*@ ensures \result == pdp_lev_permission((MacLabelCnfLevel) slev, (MacLabelCnfLevel) olev, mode);
 */
int pdp_lev_permission(PDP_LEV_T slev, PDP_LEV_T olev, int mode);

/*@ requires intersectCategories;
    requires equalCategories;
    ensures \result == pdp_cat_permission(i2m_CnfCategorySet(scat), i2m_CnfCategorySet(ocat), mode);
 */
int pdp_cat_permission(PDP_CAT_T scat, PDP_CAT_T ocat, int mode);

/*@ requires \valid(s);
    requires \valid(o);
    assigns \nothing;
    ensures \result == pdpml_conf_permission(i2m_MacLabel(s), i2m_MacLabel(o), mode);
 */
int pdpml_conf_permission(const PDPML_T *s, const PDPML_T *o, int mode);

/*@ requires ValidIntLevel(silev);
    requires ValidIntLevel(oilev);
    assigns \nothing;
    ensures \result == pdp_ilev_permission_model(i2m_IntLevel(silev), i2m_IntLevel(oilev), mode);
 */
int pdp_ilev_permission(PDP_ILEV_T silev, PDP_ILEV_T oilev, int mode);

/* mode: X_OK, R_OK, W_OK or MAY_READ, MAY_WRITE, MAY_EXEC 4 kernel */

/*@ requires \valid(s);
    requires \valid(o);
    requires ValidIntLevel(s->ilev);
    requires ValidIntLevel(o->ilev);
    assigns \nothing;
    ensures \result == (int) pdpml_permission(i2m_MacLabel(s), i2m_MacLabel(o), mode);
*/
int pdpml_permission(const PDPML_T *s, const PDPML_T *o, int mode);

#ifdef __cplusplus
}
#endif

#endif /* _PDP_COMMON_H_INCLUDED_ */
